//small assembler program for assembling programs for simulated computer


//version 2 adds labels
//labels are needed to keep track of named addresses and variables, so jump destination addresses do not need to be updated manually


#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>


using namespace std;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//STRUCT DECLARATIONS
struct SOURCE
{
	int numLines;
	int* p_Addresses; 
	int* p_NumParts;
	string** spp_ParsedData;
};

struct LUT
{
	int numEntries;	//total number of entries
	string* sp_id;	//string id by which search is done (eg labels)
	int* sp_val; //integer value found by id (eg actual address)
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCTION DECLARATIONS



//takes in asm file and replaces labels with actual addresses
SOURCE			FileToStruct( string s_source );
LUT				FindLabels ( SOURCE sParsedData ); //finds labels, makes look-up table, removes label declarations
SOURCE			SubstituteLabels ( SOURCE sParsedData, LUT luLabels ); //replaces labels with actual addresses found via LUT

//assembles prepared asm into actual machine code
string 			Assemble( SOURCE sParsedData );
unsigned int 	Convert( string* sp_parts, int chunks, int lineNumber );
unsigned int 	CharToInt ( char character );
unsigned int 	BinTextToInt ( string value );
unsigned int 	HexTextToInt ( string value );
unsigned int 	ArLogOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	ArCarryOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	CMPandShiftOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	RotOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	StackOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	MovOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	MemAccessOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
unsigned int 	JumpOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword );
//small static look-up table for named registers
string* 		SpRegNameSubstitution ( string* sp_parts, int chunks );
string			SubstituteViaLUT ( string s_operand, LUT luLabels, int operandLength ); //performs look-up, and if finds entry, substitutes it

void			ShowSource ( SOURCE sParsedData );


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main ()
{
	cout << "Assembler for 16-bit ttl computer simulation.\n\n";
	cout << "Enter the name of .asm file to convert it into hex.\n\n";
	

	
	//moving I/O into Main
	ofstream fp_out;	//"fp_out" is the name of output file stream
	ifstream fp_in;		//"fp_in" is the name of input file stream
	stringstream ss_temp; //temporary stream
	string 	s_temp, s_code,//temporary strings
			inputFile, outputFile, //names for input and output files
			s_yes,
			s_processed,
			s_tableBorder1 = "-------|----------|-----|-------|-------|-------|-------",
			s_tableBorder2 = "--------|--------|----------|----------------------------";
			
	SOURCE 	sParsedData;
	LUT		luLabels;

//	string* sp_LabelsLUT = NULL;
	
	
//load file from disk	
	cout << "Enter name of input file:\t";
	cin >> inputFile;
	inputFile += ".asm";
	const char* NameOfInputFile = inputFile.c_str();
	
	//load into stream
	fp_in.open ( NameOfInputFile, ios::in );
	ss_temp << fp_in.rdbuf();
	fp_in.close();

//saving stream into string	
	s_temp = ss_temp.str();
	
	cout << "The data saved to 's_temp' is:\n\n" << s_temp << "\n\n";

	cout << "Parsing this string...\n\n";
	
	sParsedData = FileToStruct( s_temp );

	cout << "Showing contents of SOURCE data structure \"sParsedData\":\n\n";
	ShowSource ( sParsedData );
	
//	Preprocessing steps:
//		1) make labels look-up table
//		2) remove label declaration from source

	cout << "\nCreating look-up table of labels...\n";

	luLabels = FindLabels (sParsedData);
	
	cout << "Showing contents of LUT data structure \"luLabels\":\n\n";
	cout << "Total number of entries is: " << luLabels.numEntries << "\n" << s_tableBorder2 << "\n";
	cout << " Entry# | Label  | Address\n" << s_tableBorder2 << "\n";
	for ( int i = 0; i < luLabels.numEntries; i++ )
	{
		cout << dec << i << ":\t| " << luLabels.sp_id[i] << "\t | 0x"  << setfill('0') << setw(6) << hex << luLabels.sp_val[i] << "\n";
	}
	cout << dec << s_tableBorder2 << "\n";
	
	cout << "\nPress any key to continue.\n";
	cin.get();
	cin.ignore();
	
	//display contents of data structure after labels found and their declarations removed
	
	cout << "Showing contents of SOURCE data structure \"sParsedData\" again:\n\n";
	ShowSource ( sParsedData );

//		3) replace labels with address values, where they are encountered
	
	cout << "\nSubstituting labels with actual addresses...\n";
	
	sParsedData = SubstituteLabels ( sParsedData, luLabels );
	
	cout << "Showing contents of SOURCE data structure \"sParsedData\" again:\n\n";
	ShowSource ( sParsedData );

	cout << "\nAssembling the code:\n\n";
//"Meat" of the assembler program
	s_code = Assemble( sParsedData );




	
//write file to disk	
	cout << "\n\nSave .hex file?(y/n):\t";
	cin >> s_yes;

	if ( s_yes == "y" )
	{
		cout << "\n\nEnter name of the output file:\t";
		cin >> outputFile;
		outputFile += ".hex";
		const char* NameOfOutputFile = outputFile.c_str();
			
		cout << "Contents of the new file:\n\n";
		
		cout << s_code;

		fp_out.open( NameOfOutputFile, ios::out);
		fp_out << s_code;
		fp_out.close();
			
		cout << "\n\nFile " << NameOfOutputFile << " is written.\n\n";
	}


	cout << "\nPress any key to exit the program.\n";
	cin.get();
	cin.ignore();


	return 0;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//splitting string into lines and using resulting string array
SOURCE FileToStruct( string s_source )
{
/*
struct SOURCE
{
	int numLines;
	int* p_Addresses; 
	int* p_NumParts;
	string** spp_ParsedData;
};
*/
	SOURCE 			sParsedData;
	
	stringstream	ss_source, ss_source1, ss_result;
	string			s_line, s_result, s_string, s_label;
	int 			lineCount, chunks, length, chunk;
	int*			p_NumParts = NULL, *p_Addresses = NULL;
	string**		spp_ParsedData = NULL;
	
	
	ss_source << s_source;
	
	lineCount = 0;
	while ( getline ( ss_source, s_string ) ) // getting one line at a time
	{
		if ( s_string != "" )
		{
			lineCount++;
		}
	}
	
	sParsedData.numLines = lineCount;
	
	p_NumParts = new int[lineCount];
	p_Addresses = new int[lineCount];
	spp_ParsedData = new string* [lineCount];
	
	ss_source.str("");  //clearing stream
	ss_source.clear();	//resetting stream flags
	ss_source << s_source;
	lineCount = 0;
	while ( getline ( ss_source, s_string ) ) // getting one line at a time
	{
		if ( s_string != "" )//automatically excludes empty strings
		{
			length = s_string.length(); //finding the length of line
			chunks = 0;
			for ( int i = 0; i <= length; i++ ) //going through this line character by character
			{
				if ( s_string[i] == ' ' || s_string[i] == '\t' || s_string[i] == '\0' ) //if found space, this means that chunk is over
				{
					chunks++; //counting chunks
				}
			}

			//when final count ready, make new array to fit all parts
			string* sp_parts = new string [ chunks ];
			string s_part = ""; //initialise part as empty string
			chunk = 0; //initialise chunk count with zero
			for ( int i = 0; i <= length; i++ ) //going through this line character by character
			{
				if ( s_string[i] != ' ' && s_string[i] != '\t' && s_string[i] != '\0' ) //auto exclude any whitespaces
				{
					//if non-whitespace character, add it to part
					s_part += s_string[i];
				}
				else
				{	
					//else copy s_part string to string array at index [chunk]
					if ( s_part != "" ) //part is counted and being saved only if it is a non-empty string
					{
						if ( s_part[0] == '#' )
						{	//stop going through line when comment starts
							break;
						}
						else
						{
							sp_parts[chunk] = s_part;
							s_part = "";	//re-initialise part as empty string for use in the next cycle
							chunk++; //increment chunk counter
						}
					}
				}
				
			}

			p_NumParts[lineCount] = chunk;
			p_Addresses[lineCount] = 0;//just initializing the array;no actual addresses are stored
			
			spp_ParsedData[lineCount] = sp_parts;

			lineCount++;
		}
		else
		{
			cout << "The line is empty\n";
		}
	}
	
	sParsedData.p_NumParts = p_NumParts;
	sParsedData.p_Addresses = p_Addresses;
	sParsedData.spp_ParsedData = spp_ParsedData;
	
	
	return sParsedData;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

LUT	FindLabels ( SOURCE sParsedData ) //finds labels, makes look-up table, removes label declarations
{
/*
struct SOURCE
{
	int numLines;
	int* p_Addresses; 
	int* p_NumParts;
	string** spp_ParsedData;
};
*/

/*
struct LUT
{
	int numEntries;	//total number of entries
	string* sp_id;	//id by which search is done (eg labels)
	int* sp_val; //value found by id (eg actual address)
};
*/	

	LUT luLabels;
	int length, labelCount = 0, addressCount = 0;
	stringstream ss_convert;
	string s_instr;
	
	
	
	for ( int i = 0; i < sParsedData.numLines; i++ )
	{
		length = sParsedData.spp_ParsedData[i][0].length();
		if ( sParsedData.spp_ParsedData[i][0][length-1] == ':' )//if label is found
		{
			labelCount++;
		}
	}
	
	cout << "Total of " << labelCount << " label(s) found.\n";
	
	//initializing the structure by creating appropriately sized arrays
	luLabels.numEntries = labelCount;
	luLabels.sp_id = new string[luLabels.numEntries];
	luLabels.sp_val = new int[luLabels.numEntries];
	
	labelCount = 0; //re-initialise count
	for ( int i = 0; i < sParsedData.numLines; i++ )
	{

		length = sParsedData.spp_ParsedData[i][0].length();
		if ( sParsedData.spp_ParsedData[i][0][length-1] == ':' )//if label is found
		{
			luLabels.sp_id[labelCount] = sParsedData.spp_ParsedData[i][0];
			luLabels.sp_id[labelCount].resize(length-1); //making string one character shorter
			
			//removing label declaration
			for ( int j = 0; j < (sParsedData.p_NumParts[i] - 1); j++ )
			{
				sParsedData.spp_ParsedData[i][j] = sParsedData.spp_ParsedData[i][j+1];
			}
			//decrementing part count to accomodate lesser number of parts
			sParsedData.p_NumParts[i]--;
			
			
			luLabels.sp_val[labelCount] = addressCount;
			
			
			labelCount++;
		}
		
		sParsedData.p_Addresses[i] = addressCount;
		
		s_instr = sParsedData.spp_ParsedData[i][0];
		length = s_instr.length();
		if ( s_instr == "LDi" || s_instr == "STi" || s_instr == "LDg" || s_instr == "STg" || ( s_instr[0] == 'J' && s_instr[length-1] == 'g' ) )
		{ //if two-word instruction, increment address counter by 2
			addressCount+=2;
		}
		else //one word -- increment by one
		{
			addressCount++;
		}
	}
	

	
	return luLabels;
}


SOURCE SubstituteLabels ( SOURCE sParsedData, LUT luLabels )
{
/*
struct SOURCE
{
	int numLines;
	int* p_Addresses; 
	int* p_NumParts;
	string** spp_ParsedData;
};
*/

/*
struct LUT
{
	int numEntries;	//total number of entries
	string* sp_id;	//id by which search is done (eg labels)
	int* sp_val; //value found by id (eg actual address)
};
*/	
	string s_operand, s_instr;
	int length, operandLength;

	
	
	
	for ( int i = 0; i < sParsedData.numLines; i++ )
	{
		
		
		s_instr = sParsedData.spp_ParsedData[i][0];
		length = s_instr.length();
		if ( s_instr[0] == 'J' ) //dealing with jump addresses
		{
			if ( s_instr[length-1] == 'g' ) //if global jump
			{
				operandLength = 6; //full 24-bit address, 6 hexadecimal digits
			}
			else
			{
				operandLength = 2; //only low 8 bits of address, 2 hexadecimal digits
			}
		}
		else if ( s_instr == "LDg" || s_instr == "STg" ) //direct global loads/stores
		{
			operandLength = 6; //full 24-bit address, 6 hexadecimal digits
		}
		else if ( s_instr == "LDi" || s_instr == "STi" ) //immediate loads/stores
		{
			operandLength = 4; //16-bit address inside block, 4 hexadecimal digits
		}
		else if ( s_instr == "LD" || s_instr == "ST" ) //local global loads/stores
		{
			operandLength = 2; //8-bit address inside page, 2 hexadecimal digits
		}
		else if ( s_instr == "ADD" || s_instr == "ADDC" || s_instr == "SUB" || s_instr == "SUBC" || s_instr == "AND" || s_instr == "ANDN" || s_instr == "OR" || s_instr == "ORN" || s_instr == "XOR" || s_instr == "XNOR" ) //using labels as short immediate operands for ALU instructions
		{
			operandLength = 2; //8-bit address inside page, 2 hexadecimal digits
		}
		else
		{
			operandLength = 8; //the value is not affected
		}
		
		
		for ( int j = 1; j < sParsedData.p_NumParts[i]; j++ )
		{
			s_operand = sParsedData.spp_ParsedData[i][j];
			length = s_operand.length();
			if ( s_operand[0] == '[' && s_operand[1] == '[' && s_operand[length-1] == ']' && s_operand[length-2] == ']' )
			{//get rid of parentheses
				for ( int c = 0; c < (length - 4); c++ )
				{
					s_operand[c] = s_operand[c+2];
				}
				s_operand.resize(length-4);
				
				s_operand = SubstituteViaLUT ( s_operand, luLabels, operandLength );
				s_operand = "[[" + s_operand + "]]";
			}
			else if ( s_operand[0] == '{' && s_operand[1] == '{' && s_operand[length-1] == '}' && s_operand[length-2] == '}' )
			{//get rid of parentheses
				for ( int c = 0; c < (length - 4); c++ )
				{
					s_operand[c] = s_operand[c+2];
				}
				s_operand.resize(length-4);
				
				s_operand = SubstituteViaLUT ( s_operand, luLabels, operandLength );
				s_operand = "{{" + s_operand + "}}";
			}
			else if ( s_operand[0] == '[' && s_operand[length-1] == ']' )
			{//get rid of parentheses
				for ( int c = 0; c < (length - 2); c++ )
				{
					s_operand[c] = s_operand[c+1];
				}
				s_operand.resize(length-2);
				
				s_operand = SubstituteViaLUT ( s_operand, luLabels, operandLength );
				s_operand = "[" + s_operand + "]";
			}
			else if ( s_operand[0] == '{' && s_operand[length-1] == '}' )
			{//get rid of parentheses
				for ( int c = 0; c < (length - 2); c++ )
				{
					s_operand[c] = s_operand[c+1];
				}
				s_operand.resize(length-2);
				
				s_operand = SubstituteViaLUT ( s_operand, luLabels, operandLength );
				s_operand = "{" + s_operand + "}";
			}
			else 
			{
				s_operand = SubstituteViaLUT ( s_operand, luLabels, operandLength );
			}
			
			//actual substitution
			sParsedData.spp_ParsedData[i][j] = s_operand;
		}
	}
	
	
	return sParsedData;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

string Assemble( SOURCE sParsedData )
{
	/*
struct SOURCE
{
	int numLines;
	int* p_NumParts;
	string** spp_ParsedData;
};
*/
	string s_1, s_2, s_string, s_token, s_output, s_code = "v2.0 raw\n", s_codeDisp = "";
	
	int count = 0, lcount = 0, chunks;//, chunkstart;
	string inputFile, outputFile, type, instruction;
	
	string* sp_parts;

	unsigned int codeword;

	stringstream ss_temp,
				 ss_convert, //for converting between ints to strings
				 ss_disp;
	
	

	
	
	for ( int i = 0; i < sParsedData.numLines; i++ )
	{
		
		
		sp_parts = sParsedData.spp_ParsedData[i];
		chunks = sParsedData.p_NumParts[i];
		
		
		
	//	cout << "Line " << count << " is composed of " << chunks << " chunks:\n";
		
		sp_parts = SpRegNameSubstitution ( sp_parts, chunks ); //finding named register and substitute them with regular indexes
		
		codeword = Convert( sp_parts, chunks, count );
		
		if ( (codeword & 0xffff0000) == 0 )
		{
			ss_convert << setfill('0') << setw(4) << hex << codeword << "\n";
			ss_disp << "0x" << setfill('0') << setw(6) << hex << (count + lcount) << "\t" << setfill('0') << setw(4) <<  codeword << "\t\t" << s_string << "\n";
		}
		else
		{
			ss_convert << setfill('0') << setw(4) << hex << ( (codeword & 0xffff0000) >> 16 ) << "\n";
			ss_disp << "0x" << setfill('0') << setw(6) << hex << (count + lcount) << "\t" << setfill('0') << setw(4) <<  ( (codeword & 0xffff0000) >> 16 ) << "\t";
			lcount++;
			ss_convert << setfill('0') << setw(4) << hex << (codeword & 0x0000ffff) << "\n";
			ss_disp << setfill('0') << setw(4) <<  (codeword & 0x0000ffff) << "\t" << s_string << "\n";
		}
		
		
		
		
		count++;
		delete[] sp_parts;
		sp_parts = NULL;
	}
	//fp_in.close();
	
	s_code += ss_convert.str();
	s_codeDisp = ss_disp.str();
	
	cout << "\nTotal number of lines:\t" << dec << count << "\n\n";
	cout << "resulting code:\n\n";
	cout << s_codeDisp;
	
	return s_code;
}


unsigned int CharToInt ( char character )
{
	unsigned int number;
	switch ( character )
	{
		case '0': number = 0x0000; break;
		case '1': number = 0x0001; break;
		case '2': number = 0x0002; break;
		case '3': number = 0x0003; break;
		case '4': number = 0x0004; break;
		case '5': number = 0x0005; break;
		case '6': number = 0x0006; break;
		case '7': number = 0x0007; break;
		case '8': number = 0x0008; break;
		case '9': number = 0x0009; break;
		case 'a': number = 0x000a; break;
		case 'b': number = 0x000b; break;
		case 'c': number = 0x000c; break;
		case 'd': number = 0x000d; break;
		case 'e': number = 0x000e; break;
		case 'f': number = 0x000f; break;
		default: cout << "\nCharToInt ERROR: the character is out of [0-9a-f] range!\n";
	}
		
	return number;
}

unsigned int Convert( string* sp_parts, int chunks, int lineNumber )
{
	
	unsigned int  codeword = 0x0000;
						
	cout << lineNumber << ":";
	for ( int i = 0; i < chunks; i++ )
	{
		cout << "\t" << sp_parts[i];
	}
	cout << "\n";
	
	if ( sp_parts[0] == "ADD" )
	{
		codeword |= 0x8000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "SUB" )
	{
		codeword |= 0x9000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "CMP" )
	{
		codeword |= 0x9001;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ADDC" )
	{
		codeword |= 0x8008;
		codeword = ArCarryOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "SUBC" )
	{
		codeword |= 0x9008;
		codeword = ArCarryOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "AND" )
	{
		codeword |= 0xa000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ANDN" )
	{
		codeword |= 0xb000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "OR" )
	{
		codeword |= 0xc000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ORN" )
	{
		codeword |= 0xd000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "XOR" )
	{
		codeword |= 0xe000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "XNOR" )
	{
		codeword |= 0xf000;
		codeword = ArLogOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "SHL" )
	{
		codeword |= 0xa010;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "SHR" )
	{
		codeword |= 0xb010;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ASHL" )
	{
		codeword |= 0xa014;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ASHR" )
	{
		codeword |= 0xb014;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ROLC" )
	{
		codeword |= 0xa018;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "RORC" )
	{
		codeword |= 0xb018;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ROT" )
	{
		codeword |= 0xc010;
		codeword = RotOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "INV" )
	{
		codeword |= 0xe010;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ZERO" )
	{
		codeword |= 0x8010;
		codeword = CMPandShiftOpParser( sp_parts, chunks, lineNumber, codeword );
	}

// MOV	
	else if ( sp_parts[0] == "MOV" )
	{
		codeword |= 0x1000;
		codeword = MovOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	
	
//Load / Store	
	else if ( sp_parts[0] == "LD" )
	{
		codeword |= 0x4000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ST" )
	{
		codeword |= 0x6000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "LDi" ) //immediate load, 2-word instruction
	{
		codeword |= 0x4000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "STi" ) //immediate store, 2-word instruction  ((the utility of such instruction is highly questionable))
	{
		codeword |= 0x6000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "LDg" ) //global load, 2-word instruction
	{
		codeword |= 0x4000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "STg" ) // global store, 2-word instruction
	{
		codeword |= 0x6000;
		codeword = MemAccessOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	
	
// Jumps	
	else if ( sp_parts[0] == "JMP" )
	{
		codeword |= 0x2000;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JC" )
	{
		codeword |= 0x2400;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JN" )
	{
		codeword |= 0x2200;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JZ" )
	{
		codeword |= 0x2100;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JNC" )
	{
		codeword |= 0x2300;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JNN" )
	{
		codeword |= 0x2500;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "JNZ" )
	{
		codeword |= 0x2600;
		codeword = JumpOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "NOP" ) //operation has no arguments
	{
		codeword |= 0x2700;
	}
	
// Stack operations	
	else if ( sp_parts[0] == "PUSH" )
	{
		codeword |= 0x0800;
		codeword = StackOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "POP" )
	{
		codeword |= 0x0804;
		codeword = StackOpParser( sp_parts, chunks, lineNumber, codeword );
	}
	else if ( sp_parts[0] == "ESR" )  //operation has no arguments -- it pushes PC to stack
	{
		codeword |= 0x0808;
	}
	else if ( sp_parts[0] == "RET" )  //operation has no arguments -- it pops PC from stack
	{
		codeword |= 0x080a;
	}
	
//Halt	
	else if ( sp_parts[0] == "HLT" )  //operation has no arguments
	{
		codeword |= 0x0400;
	}
	
//Data
	
	else if ( sp_parts[0][0] == '0' && sp_parts[0][1] == 'x' )
	{
		codeword = HexTextToInt( sp_parts[0] );
	}
	else
	{
		cout << "Convert Error: " << sp_parts[0] << " is not a recognized instruction, 0x0000 codeword will be returned.\n";
		codeword = 0x0000;
	}
	
	
	return codeword;
}



unsigned int BinTextToInt ( string value )
{
	int stringLength = value.length(); //
	unsigned int result =0x0;
	
	
	if ( stringLength > 1 && stringLength < 35 && value[0] == '0' && value[1] == 'b' )
	{
		for ( int i = 0; i < (stringLength-2); i++ )
		{
			
			result |= ( CharToInt( value[stringLength - 1 - i] ) << i );
		}
	}
	else
	{
		cout << "BinTextToInt Error: this is not a string with 32 bit binary value!\n";
		cout << "Length of string is: " << stringLength << ", first two characters are " << value[0] << value[1] << ";\n";
	}
		
	return result;
}

unsigned int HexTextToInt ( string value )
{
	int stringLength = value.length(); //
	unsigned int result =0x0;
	
	if ( stringLength > 1 )
	{
		if ( stringLength < 10 && value[0] == '0' && value[1] == 'x' )
		{
			for ( int i = 0; i < (stringLength-2); i++ )
			{
				
				result |= ( CharToInt( value[stringLength - 1 - i] ) << (i*4) );
			}
		}
		else if ( stringLength < 12 && value[0] == '[' && value[stringLength-1] == ']' && value[1] == '0' && value[2] == 'x' )
		{
			for ( int i = 1; i < (stringLength-3); i++ )
			{
				
				result |= ( CharToInt( value[stringLength - 1 - i] ) << ((i-1)*4) );
			}
		}
		else if ( stringLength < 12 && value[0] == '{' && value[stringLength-1] == '}' && value[1] == '0' && value[2] == 'x' )
		{
			for ( int i = 1; i < (stringLength-3); i++ )
			{
				
				result |= ( CharToInt( value[stringLength - 1 - i] ) << ((i-1)*4) );
			}
		}
		else if ( stringLength < 14 && value[0] == '[' && value[1] == '[' && value[stringLength-2] == ']' && value[stringLength-1] == ']' && value[2] == '0' && value[3] == 'x' )
		{
			for ( int i = 2; i < (stringLength-4); i++ )
			{
				
				result |= ( CharToInt( value[stringLength - 1 - i] ) << ((i-2)*4) );
			}
		}
		else if ( stringLength < 14 && value[0] == '{' && value[1] == '{' && value[stringLength-2] == '}' && value[stringLength-1] == '}' && value[2] == '0' && value[3] == 'x' )
		{
			for ( int i = 2; i < (stringLength-4); i++ )
			{
				
				result |= ( CharToInt( value[stringLength - 1 - i] ) << ((i-2)*4) );
			}
		}
		else
		{
			cout << "HexTextToInt Error: this is not a string with 8-digit hexadecimal value!\n";
			cout << "Length of string is: " << stringLength << ", first two characters are " << value[0] << value[1] << ";\n";
		}
	}
	else
	{
		cout << "HexTextToInt Error: the string is too short: only " << stringLength << " characters!\n";
	}
	
	return result;
}


unsigned int ArLogOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int 	partLength,	//lenth of value
					value;		//immediate or address value
	
	unsigned int  	regAddress; //3-bit address of 
	
	
	if ( chunks > 1 ) 
	{
		if ( sp_parts[1][0] == 'a' && sp_parts[1].length() == 2 ) //first, "destination" and/or "source 2" operand
		{
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7.\n";
			}
			if ( chunks > 2 ) 
			{
				if ( sp_parts[2][0] == 'b' && sp_parts[2].length() == 2 ) // second, "source 1" operand
				{
					regAddress = CharToInt( sp_parts[2][1] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
						
						if ( chunks > 3 )
						{
							if ( sp_parts[3][0] == 'c' ) // third, "source 2" operand
							{
								regAddress = CharToInt( sp_parts[3][1] );
								if ( regAddress > 1 && regAddress < 8 )
								{
									codeword |= regAddress;
								}
								else
								{
									cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation shold be c2...c7.\n";
									
								}
							}
							else if ( sp_parts[3][0] == '#' || sp_parts[3][0] == ' ' ) // start of comment line
							{
								//ignore, this is comment
							}
							else
							{
								cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation should start with \'c\'\n";
								cout << "Now operand is " << sp_parts[3] << "\n";
							}
						}
						
					}
					else
					{
						cout << "Error on line " << lineNumber << ": operand b for " << sp_parts[0] << " operation should be b0...b7.\n";
					}
					
				}
				else if ( sp_parts[2][0] == '0') //immediate operand, binary or hex value
				{
					codeword |= 0x0800;
					partLength = sp_parts[2].length();
					if ( partLength > 2 && sp_parts[2][1] == 'b') //binary value
					{					
						value = BinTextToInt( sp_parts[2] );
						if ( value < 256 )
						{
							codeword |= value;
						}
						else 
						{
							cout << "Error on line " << lineNumber << ": immediate value for " << sp_parts[0] << " operation should be no more than 8 bits\n";
						}
					}
					else if (  partLength > 2 && sp_parts[2][1] == 'x') //hexadecimal value
					{
						value = HexTextToInt( sp_parts[2] );
						if ( value < 256 )
						{
							codeword |= value;
						}
						else 
						{
							cout << "Error on line " << lineNumber << ": immediate value for " << sp_parts[0] << " operation should be no more than 0xff\n";
						}
					}
				}
				else
				{
					cout << "Error on line " << lineNumber << ": operands for " << sp_parts[0] << " operation are incorrect\n";
				}
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operation " << sp_parts[0] << " is missing 2nd operand -- needs at least 2 operands.\n";
			}
		}
		else
		{
			cout << "Error on line " << lineNumber << ": first operand for " << sp_parts[0] << " operation should be a0...a7;\n";
			cout << "Currently, " << sp_parts[1] << " is used instead.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation -- minimum 2 operands are needed.\n";
	}

	return codeword;
}


unsigned int ArCarryOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of 
	
	
	if ( chunks > 1 ) 
	{
		if ( sp_parts[1][0] == 'a' && sp_parts[1].length() == 2 ) //first, "destination" and/or "source 2" operand
		{
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7.\n";
			}
			if ( chunks > 2 ) 
			{
				if ( sp_parts[2][0] == 'b' && sp_parts[2].length() == 2 ) // second, "source 1" operand
				{
					regAddress = CharToInt( sp_parts[2][1] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
						
						if ( chunks > 3 )
						{
							if ( sp_parts[3][0] == 'c' ) // third, "source 2" operand
							{
								regAddress = CharToInt( sp_parts[3][1] );
								if ( regAddress > 1 && regAddress < 8 )
								{
									codeword |= regAddress;
								}
								else
								{
									cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation shold be c2...c7.\n";
									
								}
							}
							else if ( sp_parts[3][0] == '#' || sp_parts[3][0] == ' ' ) // start of comment line
							{
								//ignore, this is comment
							}
							else
							{
								cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation should start with \'c\'\n";
								cout << "Now operand is " << sp_parts[3] << "\n";
							}
						}
						
					}
					else
					{
						cout << "Error on line " << lineNumber << ": operand b for " << sp_parts[0] << " operation should be b0...b7.\n";
					}
					
				}
				//No immediate values
				else
				{
					cout << "Error on line " << lineNumber << ": operands for " << sp_parts[0] << " operation are incorrect\n";
				}
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operation " << sp_parts[0] << " is missing 2nd operand -- needs at least 2 operands.\n";
			}
		}
		else
		{
			cout << "Error on line " << lineNumber << ": first operand for " << sp_parts[0] << " operation should be a0...a7;\n";
			cout << "Currently, " << sp_parts[1] << " is used instead.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation -- minimum 2 operands are needed.\n";
	}

	return codeword;
}

unsigned int CMPandShiftOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	
	
	if ( chunks > 1 )
	{
		if ( sp_parts[1][0] == 'a' && sp_parts[1].length() == 2 ) //first, "destination" and/or "source 2" operand
		{
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7.\n";
			}
			if ( sp_parts[0] != "ZERO" )
			{
				if ( chunks > 2 )
				{
					if ( sp_parts[2][0] == 'b' && sp_parts[2].length() == 2 ) // second, "source 1" operand
					{
						regAddress = CharToInt( sp_parts[2][1] );
						if ( regAddress < 8 )
						{
							codeword |= ( regAddress << 5 );
						}
						else
						{
							cout << "Error on line " << lineNumber << ": operand b for " << sp_parts[0] << " operation should be b0...b7.\n";
						}
						//difference from ArCarryOpParser is that there is no 3rd operand
					}
					//only difference form ArLogOpParser is there could not be immediate value
					else
					{
						cout << "Error on line " << lineNumber << ": operand 2 for " << sp_parts[0] << " operation is incorrect\n";
					}
				}
				else
				{
					cout << "Error on line " << lineNumber << ": operation " << sp_parts[0] << " needs two operands.\n";
				}
			}
			
		}
		else 
		{
			cout << "Error on line " << lineNumber << ": wrong operand for " << sp_parts[0] << " operation.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation.\n";
	}
	
	return codeword;
}


unsigned int ZeroOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	
	
	if ( chunks > 1 )
	{
		if ( sp_parts[1][0] == 'a' && sp_parts[1].length() == 2 ) //first, "destination" and/or "source 2" operand
		{
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7.\n";
			}
		}
		else 
		{
			cout << "Error on line " << lineNumber << ": wrong operand for " << sp_parts[0] << " operation.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation.\n";
	}
	return codeword;
}


unsigned int RotOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register, or 4-bit
	
	
	if ( chunks > 1 && sp_parts[1][0] == 'a' && sp_parts[1].length() == 2 ) //first, "destination" and/or "source 2" operand
	{
		regAddress = CharToInt( sp_parts[1][1] );
		if ( regAddress < 8 )
		{
			codeword |= ( regAddress << 8 );
		}
		else
		{
			cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7.\n";
		}
		if ( chunks > 2 )
		{
			if ( sp_parts[2][0] == 'b' && sp_parts[2].length() == 2 ) // second, "source 1" operand
			{
				regAddress = CharToInt( sp_parts[2][1] );
				if ( regAddress < 8 )
				{
					codeword |= ( regAddress << 5 );
				}
				else
				{
					cout << "Error on line " << lineNumber << ": operand b for " << sp_parts[0] << " operation should be b0...b7.\n";
				}
				if ( chunks > 3 )
				{
					///below are different options for number presentation
					if ( sp_parts[3][0] == '0' && sp_parts[3][1] == 'x' ) // third, "rotate left by N bits" operand
					{
						regAddress = HexTextToInt ( sp_parts[3] );
						if ( regAddress < 0x10 )
						{
							codeword |= regAddress;
						}
						else
						{
							cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation should be 0x0...0xf.\n";
						}
					}
					else if ( sp_parts[3][0] == '0' && sp_parts[3][1] == 'b' ) // third, "rotate left by N bits" operand
					{
						regAddress = BinTextToInt ( sp_parts[3] );
						if ( regAddress < 0x10 )
						{
							codeword |= regAddress;
						}
						else
						{
							cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation should be 0b0000...0b1111.\n";
						}
					}
				}
				else
				{
					cout << "Error on line " << lineNumber << ": 3rd operand for " << sp_parts[0] << " operation should be single digit hexadecimal or 4-bit binary value\n";
				}
			}
			//only difference form ArLogOpParser is there could not be immediate value
			else
			{
				cout << "Error on line " << lineNumber << ": 3 operands are needed for " << sp_parts[0] << " operation, only 2 are present.\n";
			}
		}
		else
		{
			cout << "Error on line " << lineNumber << ": 3 operands are needed for " << sp_parts[0] << " operation, only one is present.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation, 3 needed.\n";
	}

	
	return codeword;
}

unsigned int StackOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	
	if ( chunks > 1 )
	{
		if ( sp_parts[0] == "POP" && ( sp_parts[1][0] == 'a' || sp_parts[1][0] == 's' ) && sp_parts[1].length() == 2 ) //"destination" operand
		{
			if ( sp_parts[1][0] == 's' ) //if special register is used, set bit4
			{
				codeword |= 0x0010;
			}
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be a0...a7 or s0...s7.\n";
			}
		}
		else if ( sp_parts[0] == "PUSH" && ( sp_parts[1][0] == 'b' || sp_parts[1][0] == 's' ) && sp_parts[1].length() == 2 ) //"source" operand
		{
			if ( sp_parts[1][0] == 's' ) //if special register is used, set bit4
			{
				codeword |= 0x0010;
			}
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 5 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": operand a for " << sp_parts[0] << " operation should be b0...b7 or s0...s7.\n";
			}
		}
		else 
		{
			cout << "Error on line " << lineNumber << ": wrong operand for " << sp_parts[0] << " operation.\n";
			cout << "Operand is: " << sp_parts[1] << ", its length is " << sp_parts[1].length() << "\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ":no operands for " << sp_parts[0] << " operation, one (name of register) is needed.\n";
	}
	
	return codeword;
}


unsigned int MovOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	
	if ( chunks > 1 )
	{
		if ( ( sp_parts[1][0] == 'a' || sp_parts[1][0] == 's' ) && sp_parts[1].length() == 2 ) //"destination" operand
		{
			if ( sp_parts[1][0] == 's' ) //if special register is used, set bit4
			{
				codeword |= 0x0010;
			}
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": first operand a for " << sp_parts[0] << " operation should be a0...a7 or s0...s7.\n";
			}
			if ( chunks > 2 )
			{
				if ( ( sp_parts[2][0] == 'b' || sp_parts[2][0] == 's' ) && sp_parts[2].length() == 2 ) //"source" operand
				{
					if ( sp_parts[2][0] == 's' ) //if special register is used, set bit4
					{
						codeword |= 0x0800;
					}
					regAddress = CharToInt( sp_parts[2][1] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
					}
					else
					{
						cout << "Error on line " << lineNumber << ": second operand a for " << sp_parts[0] << " operation should be b0...b7 or s0...s7.\n";
					}
				}
				else
				{
					cout << "Error on line " << lineNumber << ": wrong second operand (destination register name) for " << sp_parts[0] << " operation.\n";
				}
			}
			else
			{
				cout << "Error on line " << lineNumber << ": no second operand (destination register name) a for " << sp_parts[0] << " operation.\n";
			}
		}
		else 
		{
			cout << "Error on line " << lineNumber << ": wrong first operand (source register name) for " << sp_parts[0] << " operation.\n";
			cout << "Operand is: " << sp_parts[1] << ", its length is " << sp_parts[1].length() << "\n";
		}
	}
	else
	{
			cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation, it needs two operands (source and destination register names).\n";
	}
	
	
	return codeword;
}


unsigned int MemAccessOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	unsigned int  	memAddress; //8 or 24-bit address of memory location
	unsigned int 	value;		//immediate or address value
					
	if ( chunks > 1 )
	{
		if (  ( sp_parts[1][0] == 'a' || sp_parts[1][0] == 's' ) && sp_parts[1].length() == 2 ) // "destination" operand for Load
		{																									 // "source" operand for Store
			if ( sp_parts[1][0] == 's' ) //if special register is used, set bit4
			{
				codeword |= 0x0010;
			}
			regAddress = CharToInt( sp_parts[1][1] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 8 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": first operand a for " << sp_parts[0] << " operation should be a0...a7 or s0...s7.\n";
			}
			
			//check if there is a second operand (value or address)
			if (  chunks > 2 )
			{
				//"immediate value" operand
				if ( sp_parts[2][0] == '0' && sp_parts[2][1] == 'x' ) 
				{
					value = HexTextToInt( sp_parts[2] );
					if ( value <= 0xffff )
					{
						codeword = codeword << 16; //shifting instruction 16 bits left to accommodate immediate value
						codeword |= value;
					}
					else
					{
						cout << "Error on line " << lineNumber << ": second operand (immediate value) for " << sp_parts[0] << " operation could not be bigger than 0xffff.\n";
					}
				}
				
				//"immediate address in current block" operand
				//!!!Not implemented in simulation yet!!!
				else if ( sp_parts[2][0] == '[' && sp_parts[2].length() == 8 && sp_parts[2][7] == ']' )
				{
					codeword |= 0x0002; //setting "memory indirect" bit
					memAddress = HexTextToInt( sp_parts[2] );
					codeword = codeword << 16; //shifting instruction 16 bits left to accommodate immediate address
					codeword |= memAddress;
				}
				//"immediate address in block pointed to by 'Hi8' register" operand
				//!!!Not implemented in simulation yet!!!
				else if ( sp_parts[2][0] == '{' && sp_parts[2].length() == 8 && sp_parts[2][7] == '}' )
				{
					codeword |= 0x0003; //setting "memory indirect" and "MAR/Hi8" bits
					memAddress = HexTextToInt( sp_parts[2] );
					codeword = codeword << 16; //shifting instruction 16 bits left to accommodate immediate address
					codeword |= memAddress;
				}
				//"LD/ST in page"
				else if ( sp_parts[2][0] == '[' && sp_parts[2].length() == 6 && sp_parts[2][1] == '0' && sp_parts[2][2] == 'x' && sp_parts[2][5] == ']' && sp_parts[1][0] != 's' ) 
				{
					codeword |= 0x1000; //setting "direct" bit
					memAddress = HexTextToInt ( sp_parts[2] ); //converting address into int
					codeword |= memAddress; // adding low 8 bits of address to instruction
				}
				//"LD/ST global"
				else if ( sp_parts[2][0] == '[' && sp_parts[2].length() == 10 && sp_parts[2][9] == ']'  && sp_parts[1][0] != 's' ) 
				{
					codeword |= 0x1800; //setting "direct" and "global" bits
					memAddress = HexTextToInt ( sp_parts[2] ); //converting address into int
					codeword = codeword << 16; //shifting instruction 16 bits left to accommodate full address
					codeword |= memAddress; // adding full address to instruction
				}
				//"LD/ST register indirect in current block"
				else if ( sp_parts[2][0] == '[' && sp_parts[2].length() == 4 && sp_parts[2][1] == 'b' && sp_parts[2][3] == ']' ) 
				{
					codeword |= 0x0800; //setting "global/block" bit
					regAddress = CharToInt ( sp_parts[2][2] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
					}
					else
					{
						cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
					}
				}
				//"LD/ST register indirect in block pointed to by 'Hi8' register"
				else if ( sp_parts[2][0] == '{' && sp_parts[2].length() == 4 && sp_parts[2][1] == 'b' && sp_parts[2][3] == '}' ) 
				{
					codeword |= 0x0801; //setting "global/block" and "MAR/Hi8" bit
					regAddress = CharToInt ( sp_parts[2][2] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
					}
					else
					{
						cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
					}
				}
				//"LD/ST register-memory indirect in current block"
				else if ( sp_parts[2][0] == '[' && sp_parts[2].length() == 6 && sp_parts[2][1] == '[' && sp_parts[2][2] == 'b' && sp_parts[2][4] == ']' && sp_parts[2][5] == ']' ) 
				{
					cout << "[[]]";
					codeword |= 0x0802; //setting "global/block" and "mem indirect" bits
					regAddress = CharToInt ( sp_parts[2][3] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
					}
					else
					{
						cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
					}
				}
				//"LD/ST register-memory indirect in block pointed to by 'Hi8' register"
				else if ( sp_parts[2][0] == '{' && sp_parts[2].length() == 6 && sp_parts[2][1] == '{' && sp_parts[2][2] == 'b' && sp_parts[2][4] == '}' && sp_parts[2][5] == '}' ) 
				{
					cout << "{{}}";
					codeword |= 0x0803; //setting "global/block", "mem indirect" and "MAR/Hi8" bit
					regAddress = CharToInt ( sp_parts[2][3] );
					if ( regAddress < 8 )
					{
						codeword |= ( regAddress << 5 );
					}
					else
					{
						cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
					}
				}
				else 
				{
					cout << "Error on line " << lineNumber << ": wrong second operand for " << sp_parts[0] << " operation.\n";
				}
			}
			else 
			{
				cout << "Error on line " << lineNumber << ": Operation " << sp_parts[0] << " needs two operands. Only one operand is present.\n";
			}
		}
		else 
		{
			cout << "Error on line " << lineNumber << ": wrong first operand for " << sp_parts[0] << " operation.\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation.\n";
	}
	
	return codeword;
}


unsigned int JumpOpParser ( string* sp_parts, int chunks, int lineNumber, unsigned int codeword )
{
	unsigned int  	regAddress; //3-bit address of register
	unsigned int  	memAddress; //8 or 24-bit address of memory location
					
					
	if ( chunks > 1 ) // check if address to jump to is present
	{		
		//"immediate address in current block" operand
		//!!!Not implemented in simulation yet!!!
		if ( sp_parts[1][0] == '[' && sp_parts[1].length() == 8 && sp_parts[1][7] == ']' )
		{
			codeword |= 0x0002; //setting "memory indirect" bit
			memAddress = HexTextToInt( sp_parts[1] );
			codeword = codeword << 16; //shifting instruction 16 bits left to accommodate immediate address
			codeword |= memAddress;
		}
		//"immediate address in block pointed to by 'Hi8' register" operand
		//!!!Not implemented in simulation yet!!!
		else if ( sp_parts[1][0] == '{' && sp_parts[1].length() == 8 && sp_parts[1][7] == '}' )
		{
			codeword |= 0x0003; //setting "memory indirect" and "MAR/Hi8" bits
			memAddress = HexTextToInt( sp_parts[1] );
			codeword = codeword << 16; //shifting instruction 16 bits left to accommodate immediate address
			codeword |= memAddress;
		}
		//"Jump in page"
		else if ( sp_parts[1][0] == '[' && sp_parts[1].length() == 6 && sp_parts[1][5] == ']' && sp_parts[1][0] != 's' ) 
		{
			codeword |= 0x1000; //setting "direct" bit
			memAddress = HexTextToInt ( sp_parts[1] ); //converting address into int
			codeword |= memAddress; // adding low 8 bits of address to instruction
		}
		//"Jump global"
		else if ( sp_parts[1][0] == '[' && sp_parts[1].length() == 10 && sp_parts[1][9] == ']'  && sp_parts[1][0] != 's' ) 
		{
			codeword |= 0x1800; //setting "direct" and "global" bits
			memAddress = HexTextToInt ( sp_parts[1] ); //converting address into int
			codeword = codeword << 16; //shifting instruction 16 bits left to accommodate full address
			codeword |= memAddress; // adding full address to instruction
		}
		//"Jump via register indirect in current block"
		else if ( sp_parts[1][0] == '[' && sp_parts[1].length() == 4 && sp_parts[1][1] == 'b' && sp_parts[1][3] == ']' ) 
		{
			codeword |= 0x0800; //setting "global/block" bit
			regAddress = CharToInt ( sp_parts[1][2] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 5 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
			}
		}
		//"Jump via register indirect in block pointed to by 'Hi8' register"
		else if ( sp_parts[1][0] == '{' && sp_parts[1].length() == 4 && sp_parts[1][1] == 'b' && sp_parts[1][3] == '}' ) 
		{
			codeword |= 0x0801; //setting "global/block" and "MAR/Hi8" bit
			regAddress = CharToInt ( sp_parts[1][2] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 5 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
			}
		}
		//"Jump via register-memory indirect in current block"
		else if ( sp_parts[1][0] == '[' && sp_parts[1].length() == 6 && sp_parts[1][1] == '[' && sp_parts[1][2] == 'b' && sp_parts[1][4] == ']' && sp_parts[1][5] == ']' ) 
		{
			codeword |= 0x0800; //setting "global/block" bit
			regAddress = CharToInt ( sp_parts[1][3] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 5 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
			}
		}
		//"Jump via register-memory indirect in block pointed to by 'Hi8' register"
		else if ( sp_parts[1][0] == '{' && sp_parts[1].length() == 6 && sp_parts[1][1] == '{' && sp_parts[1][2] == 'b' && sp_parts[1][4] == '}' && sp_parts[1][5] == '}' ) 
		{
			codeword |= 0x0801; //setting "global/block" and "MAR/Hi8" bit
			regAddress = CharToInt ( sp_parts[1][3] );
			if ( regAddress < 8 )
			{
				codeword |= ( regAddress << 5 );
			}
			else
			{
				cout << "Error on line " << lineNumber << ": for " << sp_parts[0] << " operation register with address should have index 0 to 7.\n";
			}
		}
		else
		{
			cout << "Error on line " << lineNumber << ": wrong operand(s) for " << sp_parts[0] << " operation.\n";
			cout << "Operand is " << sp_parts[1].length() << " characters in length, and is " << sp_parts[1] << "\n";
		}
	}
	else 
	{
		cout << "Error on line " << lineNumber << ": no operands for " << sp_parts[0] << " operation, it needs one (address to jump to).\n";
	}				
	
	return codeword;
}


//following function substitutes Special Registers names with appropriate indexes starting with 's'
string* SpRegNameSubstitution ( string* sp_parts, int chunks )
{
	for ( int i = 0; i < chunks; i++ )
	{
		if ( sp_parts[i] == "Hi8" ) // High 8 bits of address -- 8bit register
		{
			sp_parts[i] = "s0";
		}
		else if ( sp_parts[i] == "IR" ) // Instruction register
		{
			sp_parts[i] = "s1";
		}
		else if ( sp_parts[i] == "SR" ) // Status register
		{
			sp_parts[i] = "s2";
		}
		else if ( sp_parts[i] == "PF" ) // 'Port Flags' register, read only
		{
			sp_parts[i] = "s3";
		}
		else if ( sp_parts[i] == "SP" ) // Stack Pointer, write only, via [Hi8][Bus16]
		{
			sp_parts[i] = "s3";
		}
		else if ( sp_parts[i] == "OP" ) // Output Parallel, write only
		{
			sp_parts[i] = "s4";
		}
		else if ( sp_parts[i] == "OS" ) // Output Serial, write only
		{
			sp_parts[i] = "s5";
		}
		else if ( sp_parts[i] == "IP" ) // Input Parallel, read only
		{
			sp_parts[i] = "s6";
		}
		else if ( sp_parts[i] == "IS" ) // Input Serial, read only
		{
			sp_parts[i] = "s7";
		}
	}
	
	
	return sp_parts;
}


string SubstituteViaLUT ( string s_operand, LUT luLabels, int operandLength )
{
	//operand length is needed for proper sizing of hexadecimal string value
	stringstream ss_convert;
	int mask;
	
	//case '0': number = 0x0000; break;
	
	switch ( operandLength )
	{
		case 1: mask = 0x0000000f; break;
		case 2: mask = 0x000000ff; break;
		case 3: mask = 0x00000fff; break;
		case 4: mask = 0x0000ffff; break;
		case 5: mask = 0x000fffff; break;
		case 6: mask = 0x00ffffff; break;
		case 7: mask = 0x0fffffff; break;
		case 8: mask = 0xffffffff; break;
		default: mask = 0x00000000; cout << "\nSubstituteViaLUT ERROR: the operandLength should be in [1-8] range!\n";
	}
	
	
	
	for ( int i = 0; i < luLabels.numEntries; i++ )
	{
		if ( s_operand == luLabels.sp_id[i] )
		{
	
			ss_convert << "0x" << setfill('0') << setw(operandLength) << hex << ( luLabels.sp_val[i] & mask );
			
			s_operand = ss_convert.str();
		
		}
	}
	
	return s_operand;
}


void ShowSource ( SOURCE sParsedData )
{
	string s_tableBorder1 = "-------|----------|-----|-------|-------|-------|-------";
	
	
	cout << "Total number of non-empty lines is: " << sParsedData.numLines << "\n" << s_tableBorder1 << "\n";
	cout << " Line# | Address  | Pt# | Parts\n" << s_tableBorder1 << "\n";
	for ( int i = 0; i < sParsedData.numLines; i++ )
	{
		cout << setfill(' ') << setw(7) << i << "| 0x" << setfill('0') << setw(6) << hex << sParsedData.p_Addresses[i] << " | " << dec << sParsedData.p_NumParts[i]  << "\t| ";
		for ( int j = 0; j < (sParsedData.p_NumParts[i] - 1); j++ )
		{
			cout << sParsedData.spp_ParsedData[i][j] << "\t ";
		}
		cout << sParsedData.spp_ParsedData[i][sParsedData.p_NumParts[i] - 1] << "\n";
	}
	cout << s_tableBorder1 << "\n\n";
	
	cout << "\nPress any key to continue.\n";
	cin.get();
	cin.ignore();
}